from pathlib import Path


class PYDEFLATE_PATHS:
    """Class to store the paths to the data and output folders."""

    package = Path(__file__).resolve().parent.parent
    data = package / "pydeflate" / ".pydeflate_data"



from pydeflate.get_data.imf_data import IMF
from pydeflate.get_data.oecd_data import OECD
from pydeflate.get_data.wb_data import WB

DATA = {
    "WEO data": IMF().update,
    "DAC1 data": OECD().update,
    "WB data": WB().update,
}


def update_all_data() -> None:
    """Run to update all underlying data."""

    for source, func in DATA.items():
        func()
        print(f"****Successfully updated {source}****\n")


if __name__ == "__main__":
    update_all_data()



__author__ = """Jorge Rivera"""
__email__ = "jorge.rivera@one.org"
__version__ = '1.3.0'


from pydeflate.deflate.deflate import deflate
from pydeflate.tools.exchange import exchange
from pydeflate.tools.update_data import update_all_data
from pydeflate.utils import warn_updates


def set_pydeflate_path(path):
    from pathlib import Path
    from pydeflate.config import PYDEFLATE_PATHS

    """Set the path to the data folder."""
    global PYDEFLATE_PATHS

    PYDEFLATE_PATHS.data = Path(path).resolve()


# check that data is fresh enough
warn_updates()



import io
import warnings
import zipfile as zf
from dataclasses import dataclass
from typing import Union

import pandas as pd
import requests
from bs4 import BeautifulSoup as Bs

from pydeflate.config import PYDEFLATE_PATHS
from pydeflate.get_data.data import Data
from pydeflate.utils import (
    base_year_dict,
    oecd_codes,
    value_index,
    update_update_date,
)

warnings.simplefilter("ignore", Warning, lineno=1013)

_BASE_URL: str = "https://stats.oecd.org/DownloadFiles.aspx?DatasetCode="
_TABLE1_URL: str = f"{_BASE_URL}TABLE1"


def _read_zip_content(request_content: bytes, file_name: str) -> pd.DataFrame:
    """Read the contents of a zip file"""
    _ = io.BytesIO(request_content)
    zip_file = zf.ZipFile(_).open(file_name)

    # Try two alternative separators to read the CSV file
    try:
        return pd.read_csv(zip_file, sep=",", encoding="ISO-8859-1", low_memory=False)

    except UnicodeDecodeError:
        return pd.read_csv(zip_file, sep="|", encoding="ISO-8859-1", low_memory=False)


def _download_bulk_file(url: str) -> bytes:
    """Get zipfile bytes from the webpage"""

    # get URL
    response = requests.get(url)

    # parse html
    page = Bs(response.text, "html.parser")
    link = list(page.find_all("a"))[0].attrs["onclick"][15:-3].replace("_", "-")
    link = f"https://stats.oecd.org/FileView2.aspx?IDFile={link}"

    return requests.get(link).content


def _fix_dac_deflator(data: pd.DataFrame):
    def __clean(group):
        mode = group.deflator.mode()[0]
        num_rows_with_mode = group.deflator.eq(mode).sum()

        if num_rows_with_mode > 3:
            group.loc[group.iso_code == "DAC", "deflator"] = mode

        return group

    data = data.groupby(["year"], group_keys=False).apply(__clean)

    return data


def _clean_dac1(df: pd.DataFrame) -> pd.DataFrame:
    """Clean DAC1 to keep only relevant information for deflators and exchange"""

    cols = {
        "DONOR": "donor_code",
        "AMOUNTTYPE": "type",
        "AIDTYPE": "aid",
        "Year": "year",
        "FLOWS": "flow",
        "Value": "value",
    }
    query = (
        "(aid == 1010 & flow == 1140 & year <2018 ) | "
        "(aid == 11010 & flow == 1160 & year >=2018)"
    )

    data = (
        df.filter(cols, axis=1)
        .rename(columns=cols)
        .astype({"year": "int"})
        .query(query)
        .filter(["donor_code", "type", "year", "value"])
        .pivot(index=["donor_code", "year"], columns=["type"], values="value")
        .reset_index()
        .assign(
            exchange=lambda d: round(d.N / d.A, 5),
            deflator=lambda d: round(100 * d.A / d.D, 6),  # implied deflator
            iso_code=lambda d: d.donor_code.map(oecd_codes),
            year=lambda d: pd.to_datetime(d.year, format="%Y"),
        )
        .assign(exchange=lambda d: d.exchange.fillna(1))
        .dropna(subset=["iso_code"])
        .filter(["iso_code", "year", "exchange", "deflator"], axis=1)
    )

    # Turning this on means matching the OECD deflator for DAC. While this would seem
    # desirable, it creates differences with the OECD's calculations for DAC Total
    # data = _fix_dac_deflator(data)

    return data


def _update_dac1() -> None:
    """Update dac1 data from OECD site and save as feather"""

    file_name = "Table1_Data.csv"

    print("Downloading DAC1 data, which may take a bit")

    zip_bytes = _download_bulk_file(url=_TABLE1_URL)
    df = _read_zip_content(request_content=zip_bytes, file_name=file_name).pipe(
        _clean_dac1
    )
    df.reset_index(drop=True).to_feather(PYDEFLATE_PATHS.data / "dac1.feather")
    print("Successfully downloaded DAC1 data")
    update_update_date("oecd_dac_data")


@dataclass
class OECD_XE(Data):
    method: Union[str, None] = "implied"
    data: pd.DataFrame = None
    """An object to download and return the latest OECD DAC exchange data"""

    def update(self, **kwargs) -> None:
        _update_dac1()

    def load_data(self, **kwargs) -> None:
        self._check_method()

        try:
            self.data = pd.read_feather(
                PYDEFLATE_PATHS.data / f"{self.available_methods()[self.method]}"
            )
        except FileNotFoundError:
            print("Data not found, downloading...")
            self.update()

        finally:
            self.data = pd.read_feather(
                PYDEFLATE_PATHS.data / f"{self.available_methods()[self.method]}"
            )

    def _get_usd_exchange(self) -> pd.DataFrame:
        if self.data is None:
            self.load_data()

        return self.data[["iso_code", "year", "exchange"]].rename(
            columns={"exchange": "value"}
        )

    def _get_exchange2usd_dict(self, currency_iso: str) -> dict:
        """Dictionary of currency_iso to USD"""

        df = self._get_usd_exchange()

        return (
            df.loc[df.iso_code == currency_iso]
            .dropna()
            .set_index("year")["value"]
            .to_dict()
        )

    def available_methods(self) -> dict:
        """The most complete dataset is obtained by deriving the exchange rate
        from the total ODA data found in Table 1"""
        return {"implied": "dac1.feather"}

    def get_data(self, currency_iso: str) -> pd.DataFrame:
        """Get an exchange rate for a given ISO"""

        df = self._get_usd_exchange()

        target_xe = self._get_exchange2usd_dict(currency_iso=currency_iso)

        df.value = df.value / df.year.map(target_xe)

        return df

    def get_deflator(self, currency_iso: str) -> pd.DataFrame:
        """get exchange rate deflator based on OECD base year and exchange rates"""

        # get exchange rates
        xe = self.get_data(currency_iso=currency_iso)

        # If currency is not valid
        if int(xe.value.sum()) == 0:
            raise ValueError(f"No currency exchange data for {currency_iso}")

        # get deflators and base year
        defl = self.data[["iso_code", "year", "deflator"]].rename(
            columns={"deflator": "value"}
        )

        base = base_year_dict(defl, "year")

        # get the exchange rate as an index based on the base year
        xe.value = value_index(xe, base)

        return xe


@dataclass
class OECD(Data):
    method: Union[str, None] = None
    data: pd.DataFrame = None
    """An object to download and return the latest OECD DAC deflators data"""

    def update(self, **kwargs) -> None:
        _update_dac1(**kwargs)

    def load_data(self, **kwargs) -> None:
        try:
            self.data = pd.read_feather(PYDEFLATE_PATHS.data / "dac1.feather")
        except FileNotFoundError:
            print("Data not found, downloading...")
            self.update()
        finally:
            self.data = pd.read_feather(PYDEFLATE_PATHS.data / "dac1.feather")

    def available_methods(self) -> dict:
        print("Only the DAC Deflators method is available")
        return {"oecd": "OECD DAC Deflator method"}

    def get_data(self) -> pd.DataFrame:
        if self.data is None:
            self.load_data()

        if self.method is not None:
            print(
                "Only the DAC Deflators method is available.\n"
                f"'Method {self.method}' ignored"
            )

        return self.data[["iso_code", "year", "deflator"]].rename(
            columns={"deflator": "value"}
        )

    def get_deflator(self, **kwargs) -> pd.DataFrame:
        xe_usd = OECD_XE().get_deflator("USA")

        df = self.get_data().merge(
            xe_usd,
            on=["iso_code", "year"],
            how="left",
            suffixes=("_def", "_xe"),
        )

        df["value"] = round(df.value_def * (df.value_xe / 100), 6)

        return df[["iso_code", "year", "value"]]



from typing import Optional, Callable, Union

import pandas as pd

from pydeflate.deflate.deflator import Deflator
from pydeflate.get_data.data import Data
from pydeflate.get_data.imf_data import IMF
from pydeflate.get_data.oecd_data import OECD, OECD_XE
from pydeflate.get_data.wb_data import WB, WB_XE
from pydeflate.utils import check_year_as_number, to_iso3, oecd_codes

DEFLATORS = {
    "oecd_dac": OECD,
    "wb": WB,
    "imf": IMF,
}

EXCHANGE = {"oecd_dac": OECD_XE, "wb": WB_XE, "imf": WB_XE}


def deflate(
    df: pd.DataFrame,
    base_year: int,
    source: str,
    method: Optional[str] = None,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    id_type: str = "ISO3",
    date_column: str = "date",
    source_col: str = "value",
    target_col: str = "value",
    to_current: bool = False,
    iso_column: Union[str, None] = None,
) -> pd.DataFrame:
    """Deflate amounts to a given currency - base year combination.

    Takes a DataFrame containing flows data and returns a DataFrame containing
    the deflated amounts. It can also convert from constant to current by specifying
    to_current as 'True'.

    Args:
        df: the DataFrame containing the flows column to be deflated. If multiple
            columns need to be deflated, the function needs to be called multiple times.
        base_year: If converting from current to constant, the target base year for the
            constant figures. If converting from constant, the base year of the data.
        source:{‘oecd_dac’, 'wb', 'imf'}
            The source of the data used to build the deflators. The value (and
            completeness) of the price deflators may change based on the source.
            Additionally, the OECD DAC data is only available for DAC donors. Both
            the IMF and WB sources use exchange rates downloaded from the World Bank.
        method:{'gdp', 'gdp_linked', 'cpi', 'pcpi', 'pcpie', None}
            The method used to calculate the price deflator:

            For World Bank (source == 'wb'):

            •'gdp': using GDP deflators.

            •'gdp_linked': a GDP deflator series which has been linked to
            produce a consistent time series to counteract breaks in series
            over time due to changes in base years, sources or methodologies.

            •'cpi': using Consumer Price Index data

            For IMF (source == 'imf'):

            •'pcpi': Consumer Price Index data

            •'pcpie': end-period Consumer Price Index (e.g. for December each year).

            For OECD DAC (source == 'oecd_dac'):
            •None

        source_currency: The iso3 code of the source currency. Note that deflators for EU countries
            are only in Euros from the year in which the Euro was adopted. To produce
            deflators only in euros, use 'emu'.

        target_currency :The iso3 code of the deflated amounts. It can be the same as the source
            currency. In cases where they are different, the exchange rate will be
            applied. To produce deflators only in euros, use 'emu'.

        id_column: The column containing the id codes (iso3 codes, for example) of the data's currency.

        id_type:The classification type for the id_column. By default, ISO3 but others are possible.
            Any options used in the Country Converter package are valid. For the OECD DAC classification, use 'DAC'.

        date_column:The column containing the date values. The column can contain years (int)
            or datetime objects.

        source_col:The column containing the data to be deflated.

        target_col: Column where the deflated data will be stored. It can be the same as the
            source column if a copy of the original isn't needed.

        to_current: If True, amounts will be treated as in constant prices and converted to current prices.

        iso_column:Provided for backwards compatibility. It is essentially an alias for id_column

    Returns:
        A pandas DataFrame containing the deflated data. Years for which there
        are no deflators will be returned as null values.


    """
    # copy the dataframe to avoid modifying the original
    df = df.copy(deep=True)

    # Backwards compatibility: if the iso_column parameter is used, reassign it to
    # the ID column
    if iso_column is not None:
        id_column = iso_column

    # keep track of original columns. This is so that the same order and columns can be
    # preserved.
    if target_col not in df.columns:
        cols = [*df.columns, target_col]
    else:
        cols = df.columns

    # Check if the source specified is part of the available deflators source
    if source not in DEFLATORS.keys():
        raise ValueError(f'source "{source}" not valid')

    # check if date format matches
    df, year_as_number = check_year_as_number(df, date_column)

    # Create ID col.  By default, if a country does not have a DAC deflator, the
    # DAC,Total deflator is used.
    if id_type == "DAC":
        df["id_"] = df[id_column].map(oecd_codes).fillna("DAC")
    else:
        df = df.pipe(
            to_iso3, codes_col=id_column, target_col="id_", src_classification=id_type
        )

    # Create exchange and price data objects. The specific objects created depend on the
    # source and/or method specified
    xe: Data = EXCHANGE[source]()
    price: Data = DEFLATORS[source](method=method)

    # Get deflator functions. The Deflator object takes in functions (as Callables)
    # instead of taking in the actual data.
    price_dfl: Callable = price.get_deflator
    x_dfl: Callable = xe.get_deflator

    # Get currency exchange DataFrame. This is based on the target currency
    x_rate: pd.DataFrame = xe.get_data(
        currency_iso=target_currency if not to_current else source_currency
    ).copy(deep=True)

    # Create a Deflator object and get the deflator DataFrame.
    deflator = (
        Deflator(
            base_year=base_year,
            xe_deflator=x_dfl,
            price_deflator=price_dfl,
            xe=x_rate,
            source_currency=source_currency if not to_current else target_currency,
            target_currency=target_currency if not to_current else source_currency,
        )
        .deflator()
        .rename(columns={"iso_code": "id_", "year": date_column})
    )

    # Merge the original data with the deflator DataFrame
    df = df.merge(
        deflator,
        on=["id_", date_column],
        how="left",
        suffixes=("", "_"),
    )

    # If the reverse option is used, multiply the data. Else divide.
    if to_current:
        df[target_col] = df[source_col] * (df.deflator / 100)

    else:
        df[target_col] = df[source_col] / (df.deflator / 100)

    # Keep only the columns present in the original DataFrame (including order)
    df = df.filter(cols, axis=1)

    # If the year is passed as a number, convert it back to a number.
    if year_as_number:
        df[date_column] = df[date_column].dt.year

    return df



import datetime
import json
import warnings
from typing import Union

import country_converter as coco
import numpy as np
import pandas as pd

from pydeflate import config

CC = coco.CountryConverter()


def _diff_from_today(date: datetime.datetime):
    """Compare to today"""

    today = datetime.datetime.today()

    return (today - date).days


def warn_updates():

    with open(config.PYDEFLATE_PATHS.data / "data_updates.json") as file:
        updates = json.load(file)

    for source, date in updates.items():
        d = datetime.datetime.strptime(date, "%Y-%m-%d")
        if _diff_from_today(d) > 50:
            message = (
                f'\n\nThe underlying data for "{source}" has not been updated'
                f" in over {_diff_from_today(d)} days. \nIn order to use"
                " pydeflate with the most recent data, please run:\n"
                "`pydeflate.update_all_data()`"
            )
            warnings.warn(message)


def update_update_date(source: str):
    """Update the most recent update date for data to today"""

    today = datetime.datetime.today().strftime("%Y-%m-%d")

    # Check to see if specified path contains an update file. Create one if not
    if not (config.PYDEFLATE_PATHS.data / "data_updates.json").exists():
        updates = {}
        with open(config.PYDEFLATE_PATHS.data / "data_updates.json", "w") as outfile:
            json.dump(updates, outfile)

    with open(config.PYDEFLATE_PATHS.data / "data_updates.json") as file:
        updates = json.load(file)

    updates[source] = today

    with open(config.PYDEFLATE_PATHS.data / "data_updates.json", "w") as outfile:
        json.dump(updates, outfile)


oecd_codes = {
    1: "AUT",
    2: "BEL",
    3: "DNK",
    4: "FRA",
    5: "DEU",
    6: "ITA",
    7: "NLD",
    8: "NOR",
    9: "PRT",
    10: "SWE",
    11: "CHE",
    12: "GBR",
    18: "FIN",
    20: "ISL",
    21: "IRL",
    22: "LUX",
    40: "GRC",
    50: "ESP",
    61: "SVN",
    68: "CZE",
    69: "SVK",
    75: "HUN",
    76: "POL",
    301: "CAN",
    302: "USA",
    701: "JPN",
    742: "KOR",
    801: "AUS",
    820: "NZL",
    918: "EUI",
    30: "CYP",
    45: "MLT",
    55: "TUR",
    62: "HRV",
    70: "LIE",
    72: "BGR",
    77: "ROU",
    82: "EST",
    83: "LVA",
    84: "LTU",
    87: "RUS",
    130: "DZA",
    133: "LBY",
    358: "MEX",
    543: "IRQ",
    546: "ISR",
    552: "KWT",
    561: "QAT",
    566: "SAU",
    576: "ARE",
    611: "AZE",
    613: "KAZ",
    732: "TWN",
    764: "THA",
    765: "TLS",
    20001: "DAC",
    20002: "MUL",
    20003: "G7C",
    20006: "NDA",
}

emu = [
    "AUT",
    "BEL",
    "CYP",
    "EST",
    "FIN",
    "FRA",
    "DEU",
    "GRC",
    "IRL",
    "ITA",
    "LVA",
    "LTU",
    "LUX",
    "MLT",
    "NLD",
    "PRT",
    "SVK",
    "SVN",
    "ESP",
]


def clean_number(number):
    """Clean a number and return as float"""
    import re

    if not isinstance(number, str):
        number = str(number)

    number = re.sub(r"[^\d.]", "", number)

    if number == "":
        return np.nan

    return float(number)


def base_year_dict(df: pd.DataFrame, date_col: str = "date") -> dict:
    """Return dictionary of base years by iso_code"""

    return df.loc[df.value == 100].set_index("iso_code")[date_col].to_dict()


def value_index(df: pd.DataFrame, base_dict: dict) -> pd.Series:
    """Reindex a series given a base year dictionary"""

    data = df.copy()
    df_ = df.copy()

    df_["base"] = df_.iso_code.map(base_dict)
    df_ = df_.loc[df_.year == df_.base]

    base_values = (
        df_.loc[lambda d: d.value.notna()]
        .round(6)
        .set_index("iso_code")["value"]
        .to_dict()
    )

    return round(100 * data.value / data.iso_code.map(base_values), 6)


def rebase(df_: pd.DataFrame, base_year: int) -> pd.Series:
    """Rebase values to a given base year"""

    base_values = (
        df_.loc[df_.year.dt.year == base_year].set_index("iso_code")["value"].to_dict()
    )

    return round(100 * df_.value / df_.iso_code.map(base_values), 6)


def check_method(method: str, methods: dict):
    """Check whether a given method is in a methods dictionary"""

    if method not in methods.keys():
        raise ValueError(f'Method "{method}" not valid. Please see documentation.')


def check_year_as_number(df: pd.DataFrame, date_column: str) -> (pd.DataFrame, bool):
    """Check whether the date column contains an int instead of datetime.
    This changes the column to datetime and returns a flag"""

    if pd.api.types.is_numeric_dtype(df[date_column]):
        df[date_column] = pd.to_datetime(df[date_column], format="%Y")
        year_as_number = True
    else:
        year_as_number = False

    return df, year_as_number


def to_iso3(
    df: pd.DataFrame,
    codes_col: str,
    target_col: str,
    src_classification: Union[str, None] = None,
    not_found: Union[str, None] = None,
) -> pd.DataFrame:
    df[target_col] = CC.convert(
        df[codes_col], src=src_classification, to="ISO3", not_found=not_found
    )
    return df



from dataclasses import dataclass
from typing import Union, Callable

import pandas as pd
from weo import WEO, all_releases, download

from pydeflate import utils
from pydeflate.config import PYDEFLATE_PATHS
from pydeflate.get_data.data import Data


def _check_weo_parameters(
    latest_y: Union[int, None] = None, latest_r: Union[int, None] = None
) -> (int, int):
    """Check parameters and return max values or provided input"""
    if latest_y is None:
        latest_y = max(*all_releases())[0]

    # if latest release isn't provided, take max value
    if latest_r is None:
        latest_r = max(*all_releases())[1]

    return latest_y, latest_r


def _update_weo(latest_y: int = None, latest_r: int = None) -> None:
    """Update data from the World Economic Outlook, using WEO package"""

    latest_y, latest_r = _check_weo_parameters(latest_y, latest_r)

    # Download the file from the IMF website and store in directory
    download(
        latest_y,
        latest_r,
        directory=PYDEFLATE_PATHS.data,
        filename=f"weo{latest_y}_{latest_r}.csv",
    )
    utils.update_update_date("imf")


@dataclass
class IMF(Data):
    method: Union[str, None] = None
    data: pd.DataFrame = None
    """An object to download and return the latest IMF WEO data for several indicators"""

    def update(
        self, latest_y: Union[int, None] = None, latest_r: Union[int, None] = None
    ) -> None:
        """Update the stored WEO data, using WEO package.

        Args:
            latest_y: passed only optional to override the behaviour to get the latest
            release year for the WEO.
            latest_r: passed only optionally to override the behaviour to get the latest
            released value (1 or 2).
        """
        _update_weo(latest_y=latest_y, latest_r=latest_r)

    def load_data(
        self, latest_y: Union[int, None] = None, latest_r: Union[int, None] = None
    ) -> None:
        """loading WEO as a clean dataframe

        Args:
            latest_y: passed only optional to override the behaviour to get the latest
            release year for the WEO.
            latest_r: passed only optionally to override the behaviour to get the latest
            released value (1 or 2).
        """

        latest_y, latest_r = _check_weo_parameters(latest_y, latest_r)

        names = {
            "ISO": "iso_code",
            "WEO Subject Code": "indicator",
            "Subject Descriptor": "indicator_name",
            "Units": "units",
            "Scale": "scale",
        }
        to_drop = [
            "WEO Country Code",
            "Country",
            "Subject Notes",
            "Country/Series-specific Notes",
            "Estimates Start After",
        ]
        try:
            df = WEO(PYDEFLATE_PATHS.data / f"weo{latest_y}_{latest_r}.csv").df

        except FileNotFoundError:
            print("No IMF WEO data found, downloading latest release")
            self.update()

        finally:
            df = WEO(PYDEFLATE_PATHS.data / f"weo{latest_y}_{latest_r}.csv").df

        self.data = (
            df.drop(to_drop, axis=1)
            .rename(columns=names)
            .melt(id_vars=names.values(), var_name="date", value_name="value")
            .assign(
                year=lambda d: pd.to_datetime(d.date, format="%Y"),
                value=lambda d: d.value.apply(utils.clean_number),
            )
            .dropna(subset=["value"])
            .drop("date", axis=1)
            .reset_index(drop=True)
        )

    def _get_indicator(self, indicator: str) -> pd.DataFrame:
        """Get a specified imf indicator from the downloaded WEO file"""

        return (
            self.data.loc[lambda d: d.indicator == indicator]
            .filter(["iso_code", "year", "value"], axis=1)
            .sort_values(["iso_code", "year"])
            .reset_index(drop=True)
        )

    def inflation_acp(self) -> pd.DataFrame:
        """Indicator PCPI, Index"""
        return self._get_indicator("PCPI")

    def inflation_epcp(self) -> pd.DataFrame:
        """Indicator PCPIE, Index"""
        return self._get_indicator("PCPIE")

    def inflation_gdp(self) -> pd.DataFrame:
        return self._get_indicator("NGDP_D")

    def available_methods(self) -> dict[str, Callable]:
        return {
            "gdp": self.inflation_gdp,
            "pcpi": self.inflation_acp,
            "pcpie": self.inflation_epcp,
        }

    def get_data(self, **kwargs):
        pass

    def get_deflator(self, **kwargs) -> pd.DataFrame:
        """Get the deflator DataFrame for the specified method"""

        self._check_method()

        if self.data is None:
            self.load_data()

        return self.available_methods()[self.method]()



"""Defines the deflator class"""

from dataclasses import dataclass
from typing import Union, Callable

import numpy as np
import pandas as pd

from pydeflate import utils


@dataclass
class Deflator:
    base_year: int
    xe_deflator: Callable
    price_deflator: Callable
    xe: pd.DataFrame
    source_currency: str = "LCU"
    target_currency: str = "USA"
    deflator_df: Union[None, pd.DataFrame] = None

    """A class to produce deflators based on the given parameters.
    
    Deflators take into account changes in prices and the evolution of currency exchange
    values. They compensate for both factors in order to produce a standard unit of 
    measurement, regardless of the year of the flow.
    
    Args:
        base_year: the year in which the delfator equals 100 (if the source and target
        currency are equal).
        xe_deflator: the exchange rate deflator function to be used.
        price_deflator: the price deflator function to be used.
        xe: exchange rate information used in producing the deflators
        source_currency: the ISO3 code of the country which uses the source currency
        target_currency: the ISO3 code of the country which used the target currency
        deflator_df: the resulting DataFrame containing the deflators data.
    """

    def __align_currencies(
        self,
        deflator: pd.DataFrame,
        xe: pd.DataFrame,
    ) -> pd.DataFrame:
        """Check whether an exchange rate needs to be applied to a deflator set
        in order to properly convert data from source to target currency"""

        # If both the source and target currency match, do nothing
        if self.source_currency == self.target_currency:
            return deflator

        if self.source_currency == "LCU":
            deflator = deflator.merge(xe, on=["iso_code", "year"], suffixes=("", "_xe"))
            deflator.value = deflator.value * deflator.value_xe
        else:
            xe = xe.loc[xe.iso_code == self.source_currency]
            deflator = deflator.merge(
                xe.drop("iso_code", axis=1), on=["year"], suffixes=("", "_xe")
            )
            deflator.value = deflator.value * deflator.value_xe

        # Check if data is valid
        if int(deflator.value_xe.sum()) == 0:
            raise ValueError(f"Data for {self.source_currency} is not valid")

        return deflator.drop("value_xe", axis=1)

    def _get_and_deflate(self, func: Callable):
        if self.target_currency is None:
            df = func()
        else:
            df = func(currency_iso=self.target_currency)

        if self.base_year not in set(df.year.dt.year):
            raise ValueError(f"Base year ({self.base_year} not valid.")

        df["value"] = utils.rebase(df, self.base_year)

        return df

    def _defl_pipeline(
        self,
    ) -> pd.DataFrame:
        """Returns a dataframe with price-currency deflators."""

        # get exchange rate and rebase
        xed = self._get_and_deflate(func=self.xe_deflator)

        # get price deflators and rebase
        prd = self._get_and_deflate(func=self.price_deflator)

        df = prd.merge(
            xed, on=["iso_code", "year"], how="left", suffixes=("_def", "_xe")
        )

        df["value"] = round(100 * df.value_def / df.value_xe, 6)

        return (
            df.filter(["iso_code", "year", "value"], axis=1)
            .assign(value=lambda d: d.value.replace(0, np.nan))
            .dropna(subset=["value"])
            .reset_index(drop=True)
            .pipe(self.__align_currencies, xe=self.xe)
            .rename(columns={"value": "deflator"})
        )

    def _set_deflator_df(self, df: pd.DataFrame) -> None:
        self.deflator_df = df

    def deflator(self) -> pd.DataFrame:
        """Execute the pipeline to create the deflators. It uses the information
        passed when creating the Deflator object"""
        self._defl_pipeline().pipe(self._set_deflator_df)

        return self.deflator_df



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union

import pandas as pd


@dataclass
class Data(ABC):
    method: Union[str, None]
    data: pd.DataFrame
    """Abstract class defining the basic structure and functionality of Data classes.
    Data classes store the price or exchange data from different sources."""

    @abstractmethod
    def update(self, **kwargs) -> None:
        """Update underlying data"""
        pass

    @abstractmethod
    def load_data(self, **kwargs) -> None:
        """Load required data to construct deflator"""
        pass

    @abstractmethod
    def available_methods(self) -> dict:
        """Return a dictionary of available methods with their functions"""
        pass

    def _check_method(self) -> None:
        if self.method is None:
            raise AttributeError(
                f"`method` must be defined: " f"{list(self.available_methods().keys())}"
            )

    def set_method(self, method: str) -> Data:
        self.method = method
        return self

    @abstractmethod
    def get_data(self, **kwargs) -> pd.DataFrame:
        pass

    @abstractmethod
    def get_deflator(self, **kwargs) -> pd.DataFrame:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydeflate', 'pydeflate.deflate', 'pydeflate.get_data', 'pydeflate.tools']

package_data = \
{'': ['*'], 'pydeflate': ['.pydeflate_data/*']}

install_requires = \
['BeautifulSoup4>=4.11.1,<5.0.0',
 'country-converter>=0.8.0,<0.9.0',
 'numpy>=1.24.1,<2.0.0',
 'pandas-datareader>=0.10.0,<0.11.0',
 'pandas>=1.5.2,<2.0.0',
 'pyarrow>=11,<12',
 'requests>=2.28.1,<3.0.0',
 'weo>=0.7.4,<0.8.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'pydeflate',
    'version': '1.3.0',
    'description': 'Package to convert current prices figures to constant prices and vice versa',
    'long_description': 'The pydeflate Package\n=====================\n\n|pypi|\n|Documentation Status|\n|black|\n|Downloads|\n|Coverage|\n\n\n**Pydeflate** is a Python package to convert flows data to constant\nprices. This can be done from any source currency to any desired base\nyear and currency. **Pydeflate** can also be used to convert constant\ndata to current prices and to convert from one currency to another (in\ncurrent and constant prices). Users can choose the source of the\nexchange and deflator/prices data (IMF, World Bank or OECD DAC).\n\n-  Free software: MIT license\n-  Documentation: https://pydeflate.readthedocs.io.\n\nInstallation\n------------\n\npydeflate can be installed from PyPI. From the command line:\n\n::\n\n   pip install pydeflate --upgrade\n\nAlternatively, the source code is available on\n`GitHub <https://github.com/jm-rivera/pydeflate>`__.\n\nUsage\n-----\n\nBasic usage\n~~~~~~~~~~~\n\nConvert data expressed in current USD prices to constant EUR prices for\na given base year:\n\n::\n\n   import pydeflate\n   import pandas as pd\n\n   # example data\n   data = {\'iso_code\': [\'FRA\',\'USA\', \'GTM\'],\n           \'year\': [2017, 2017, 2017],\n           \'value\': [50, 100, 200]}\n\n   # create an example dataframe, in current USD prices\n   df = pd.DataFrame.from_dict(data)\n\n   # convert to EUR 2015 constant prices\n   df_constant = pydeflate.deflate(\n       df = df,\n       base_year = 2015,\n       source = \'wb\',\n       method = \'gdp\',\n       source_currency = "USA", # since data is in USD\n       target_currency = "EMU", # we want the result in constant EUR\n       id_column = "iso_code",\n       id_type = "ISO3", # specifying this is optional in most cases\n       date_column = "year",\n       source_col = "value",\n       target_col = "value_constant",\n   ) \n           \n   print(df_constant)\n\nThis results in a dataframe containing a new column ``value_constant``\nin 2015 constant prices. In the background, pydeflate takes into\naccount:\n\n-  changes in princes, through a gdp deflator in this case\n-  changes in exchange rates overtime\n\nPydeflate can also handle data that is expressed in local currency\nunits. In that case, users can specify ``LCU`` as the source currency.\n\n::\n\n   import pydeflate\n   import pandas as pd\n\n   #example data\n   data = {\'country\': [\'United Kingdom\',\'United Kingdom\', \'Japan\'],\n           \'date\': [2011, 2015, 2015],\n           \'value\': [100, 100, 100]}\n\n   #create an example dataframe, in current local currency units \n   df = pd.DataFrame.from_dict(data)\n\n   #convert to USD 2018 constant prices\n   df_constant = pydeflate.deflate(\n       df = df,\n       base_year = 2018,\n       source = \'imf\',\n       method = \'pcpi\',\n       source_currency = "LCU", #since data is in LCU\n       target_currency = "USA", #to get data in USD\n       id_column = "iso_code",\n       date_column = "date",\n       source_col = "value",\n       target_col = "value", #to not create a new column\n   ) \n           \n   print(df_constant)\n\nUsers can also convert a dataset expressed in constant prices to current\nprices using pydeflate. To avoid introducing errors, users should know\nwhich methodology/ data was used to create constant prices by the\noriginal source. The basic usage is the same as before, but the\n``to_current`` parameter is set to ``True``.\n\nFor example, to convert DAC data expressed in 2016 USD constant prices\nto current US dollars:\n\n::\n\n   import pydeflate\n   import pandas as pd\n\n   #example data\n   data = {\'dac_code\': [302, 4, 4],\n           \'date\': [2010, 2016, 2018],\n           \'value\': [100, 100, 100]}\n\n   #create an example dataframe, in current local currency units \n   df = pd.DataFrame.from_dict(data)\n\n   #convert to USD 2018 constant prices\n   df_current = pydeflate.deflate(\n       df = df,\n       base_year = 2016,\n       source = \'oecd_dac\', \n       source_currency = "USA", #since data is in USD constant\n       target_currency = "LCU", #to get the current LCU figures\n       id_column = "dac_code",\n       id_type = "DAC",\n       date_column = "date",\n       source_col = "value",\n       target_col = "value_current", \n       to_current = True,   \n   ) \n           \n   print(df_current)\n\nData source and method options\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nA ``source`` and a ``method`` for the exchange and price/gdp deflators\nmust be chosen. The appropriate combination depends on the objectives of\nthe project or the nature of the original data.\n\nIn terms of price or GDP deflators, pydeflate provides the following\n``methods``:\n\n-  World Bank (“wb”):\n\n   -  ``gdp``: in order to use GDP deflators.\n   -  ``gdp_linked``: to use the World Bank’s GDP deflator series which\n      has been linked to produce a consistent time series to counteract\n      breaks in series over time due to changes in base years, sources\n      or methodologies.\n   -  ``cpi``: to use Consumer Price Index data\n\n-  International Monetary Fund World Economic Outlook (“imf”):\n\n   -  ``gdp``: in order to use GDP deflators.\n   -  ``pcpi``: in order to use Consumer Price Index data.\n   -  ``pcpie``: to use end-of-period Consumer Price Index data\n      (e.g.\xa0for December each year).\n\n-  OECD Development Assistance Committee (“oecd_dac”):\n\n   -  ``None``: for consistency with how the DAC calculates deflators,\n      only their methodology is accepted/used with this data.\n\nThe source of the exchange rate data depends on the source selected.\nBoth ``imf`` and ``wb`` use data from the International Monetary Fund\n(``LCU per US$, yearly average``). The OECD Development Assistance\nCommittee data uses different exchange rates. When ``oecd_dac`` is\nselected as the source, the OECD DAC exchange rates (``LCU per US$``)\nare used. Exchange rates between two non USD currency pairs are derived\nfrom the LCU to USD exchange rates selected.\n\nAdditional features\n~~~~~~~~~~~~~~~~~~~\n\nPydeflate relies on data from the World Bank, IMF and OECD for its\ncalculations. This data is updated periodically. If the version of the\ndata stored in the user’s computer is older than 50 days, pydeflate will\nshow a warning on import.\n\nUsers can always update the underlying data by using:\n\n::\n\n   import pydeflate\n\n   pydeflate.update_all_data()\n\nPydeflate also provides users with a tool to exchange figures from one\ncurrency to another, without applying any deflators. This should only be\nused on numbers expressed in current prices, however.\n\nFor example, to convert numbers in current Local Currency Units (LCU) to\ncurrent Canadian Dollars:\n\n::\n\n   import pydeflate\n   import pandas as pd\n\n   #example data\n   data = {\'iso_code\': [\'GBR\',\'CAN\', \'JPN\'],\n           \'date\': [2011, 2015, 2015],\n           \'value\': [100, 100, 100]}\n\n   #create an example dataframe, in current local currency units \n   df = pd.DataFrame.from_dict(data)\n\n   #convert to USD 2018 constant prices\n   df_can = pydeflate.exchange(\n       df = df,\n       source_currency = "LCU", #since data is in LCU\n       target_currency = "CAN", #to get data in Canadian Dollars\n       rates_source = \'wb\', #this is the same as IMF exchange rates\n       value_column = \'value\',\n       target_column = \'value_CAN\',\n       id_column = "iso_code",\n       id_type = "ISO3"\n       date_column = "date",\n   ) \n           \n   print(df_can)\n\nCredits\n-------\n\nThis package relies on data from the following sources:\n\n-  OECD DAC: https://www.oecd.org/dac/ (Official Development assistance\n   data (DAC1), DAC deflators, and exchange rates used by the DAC)\n-  IMF World Economic Outlook: https://www.imf.org/en/Publications/WEO\n   (GDP and price deflators)\n-  World Bank DataBank: https://databank.worldbank.org/home.aspx\n   (exchange rates, GDP and price deflators)\n\nThis data is provided based on the terms and conditions set by the\noriginal sources\n\n.. |pypi| image:: https://img.shields.io/pypi/v/pydeflate.svg\n   :target: https://pypi.python.org/pypi/pydeflate\n.. |Documentation Status| image:: https://readthedocs.org/projects/pydeflate/badge/?version=latest\n   :target: https://pydeflate.readthedocs.io/en/latest/?version=latest\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n.. |Downloads| image:: https://pepy.tech/badge/pydeflate/month\n   :target: https://pepy.tech/project/pydeflate\n\n.. |Coverage| image:: https://codecov.io/gh/jm-rivera/pydeflate/branch/main/graph/badge.svg?token=uwKI5DyO3w\n   :target: https://codecov.io/gh/jm-rivera/pydeflate\n\n\nGbemisola Joel-Osoba provided extensive feedback and testing of version 1.\n',
    'author': 'Jorge Rivera',
    'author_email': 'jorge.rivera@one.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jm-rivera/pydeflate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Union

import pandas as pd
from pandas_datareader import wb

from pydeflate import config
from pydeflate.get_data.data import Data
from pydeflate.utils import emu, value_index, update_update_date

_INDICATORS: dict = {
    "gdp": "NY.GDP.DEFL.ZS",
    "gdp_linked": "NY.GDP.DEFL.ZS.AD",
    "cpi": "FP.CPI.TOTL",
    "exchange": "PA.NUS.FCRF",
    "effective_exchange": "PX.REX.REER",
}

START: int = 1950
END: int = 2025


def _download_wb_indicator(
    indicator: str, start: int, end: int, dev: bool = False
) -> None:
    """Download an indicator from WB (caching if applicable)"""

    df = (
        wb.WorldBankReader(
            symbols=indicator,
            countries="all",
            start=start,
            end=end,
        )
        .read()
        .reset_index(drop=False)
    )

    countries = wb.get_countries()
    countries[["name", "iso3c"]].to_csv(
        config.PYDEFLATE_PATHS.data / f"wb_class.csv", index=False
    )

    df.reset_index(drop=True).to_feather(
        config.PYDEFLATE_PATHS.data / f"{indicator}_{start}_{end}.feather"
    )
    print(f"Successfully updated {indicator} for {start}-{end}")
    update_update_date("wb")


def _get_iso3c():
    countries = pd.read_csv(config.PYDEFLATE_PATHS.data / f"wb_class.csv")
    return countries[["name", "iso3c"]].set_index("name")["iso3c"].to_dict()


def _clean_wb_indicator(
    data: pd.DataFrame,
    indicator: str,
) -> pd.DataFrame:
    """Add iso_code, change value name to value and sort"""

    return (
        data.assign(
            iso_code=lambda d: d.country.map(_get_iso3c()),
            year=lambda d: pd.to_datetime(d.year, format="%Y"),
        )
        .rename(columns={indicator: "value"})
        .dropna(subset=["iso_code"])
        .sort_values(["iso_code", "year"])
        .reset_index(drop=True)
        .filter(["year", "iso_code", "value"], axis=1)
    )


@dataclass
class WB_XE(Data):
    method: Union[str, None] = "exchange"
    data: pd.DataFrame = None

    def update(self, **kwargs) -> None:
        _ = [
            _download_wb_indicator(i, START, END)
            for i in self.available_methods().values()
        ]

    def load_data(self, **kwargs) -> None:
        self._check_method()
        indicator = _INDICATORS[self.method]

        try:
            self.data = pd.read_feather(
                config.PYDEFLATE_PATHS.data / f"{indicator}_{START}_{END}.feather"
            ).pipe(_clean_wb_indicator, indicator)
        except FileNotFoundError:
            print("Data not found. Updating...")
            self.update()
        finally:
            self.data = pd.read_feather(
                config.PYDEFLATE_PATHS.data / f"{indicator}_{START}_{END}.feather"
            ).pipe(_clean_wb_indicator, indicator)

    def available_methods(self) -> dict:
        return {
            k: v
            for k, v in _INDICATORS.items()
            if k in ["exchange", "effective_exchange"]
        }

    def _get_usd_exchange(self) -> pd.DataFrame:
        """Official exchange rate refers to the exchange rate determined by
        national authorities or to the rate determined in the legally
        sanctioned exchange market. It is calculated as an annual average based on
        monthly averages (local currency units relative to the U.S. dollar)."""

        # get exchange rates
        if self.data is None:
            self.load_data()

        df = self.data

        eur = df.loc[df.iso_code == "EMU"].dropna().set_index("year")["value"].to_dict()

        # Euro area countries without exchange rates
        eur_mask = (df.iso_code.isin(emu)) & (df.value.isna())

        # Assign EURO exchange rate to euro are countries from year euro adopted
        df.loc[eur_mask, "value"] = df.year.map(eur)

        return df

    def _get_exchange2usd_dict(self, currency_iso: str) -> dict:
        """Dictionary of currency_iso to USD"""

        df = self._get_usd_exchange()

        return (
            df.loc[df.iso_code == currency_iso]
            .dropna()
            .set_index("year")["value"]
            .to_dict()
        )

    def _get_euro2usd(self) -> dict:
        """Dictionary of EUR to USD exchange rates"""

        return self._get_exchange2usd_dict(currency_iso="EMU")

    def _get_can2usd(self) -> dict:
        """Dictionary of CAN to USD exchange rates"""

        return self._get_exchange2usd_dict(currency_iso="CAN")

    def _get_gbp2usd(self) -> dict:
        """Dictionary of GBP to USD exchange rates"""

        return self._get_exchange2usd_dict(currency_iso="GBR")

    def get_data(self, currency_iso: str) -> pd.DataFrame:
        """Get exchange rates based on a given currency/country (from LCU)"""
        # Get WB exchange rates
        df = self._get_usd_exchange()
        target_xe = self._get_exchange2usd_dict(currency_iso=currency_iso)

        df.value = df.value / df.year.map(target_xe)

        return df

    def get_deflator(
        self, currency_iso: str = "USA", base_year: int = 2010
    ) -> pd.DataFrame:
        from datetime import datetime

        # get exchange rates
        xe = self.get_data(currency_iso=currency_iso)

        # get deflators and base year
        base = {iso: datetime(base_year, 1, 1) for iso in xe.iso_code.unique()}

        # get the exchange rate as an index based on the base year
        xe.value = value_index(xe, base)

        return xe


@dataclass
class WB(Data):
    method: Union[str, None] = None
    data: pd.DataFrame = None
    """An object to download and return the latest WB exchange and price data"""

    def update(self, **kwargs) -> None:
        """Update data for all WB indicators"""

        _ = [_download_wb_indicator(i, START, END) for i in _INDICATORS.values()]

    def load_data(
        self,
        indicator: str,
    ) -> None:
        self.data = pd.read_feather(
            config.PYDEFLATE_PATHS.data / f"{indicator}_{START}_{END}.feather"
        ).pipe(_clean_wb_indicator, indicator)

    def available_methods(self) -> dict:
        return {
            k: v
            for k, v in _INDICATORS.items()
            if k not in ["exchange", "effective_exchange"]
        }

    def _get_indicator(self) -> pd.DataFrame:
        self._check_method()
        indicator = _INDICATORS[self.method]

        if self.data is None:
            self.load_data(indicator=indicator)

        return self.data

    def get_data(self):
        raise NotImplementedError

    def get_deflator(self, **kwargs) -> pd.DataFrame:
        return self._get_indicator()
